from dune.codegen.generation import (cached,
                                     class_member,
                                     get_counted_variable,
                                     global_context,
                                     include_file,
                                     preamble,
                                     )
from dune.codegen.pdelab.driver import (FEM_name_mangling,
                                        get_form_ident,
                                        get_trial_element,
                                        is_stationary,
                                        preprocess_leaf_data,
                                        )
from dune.codegen.pdelab.driver.driverblock import (name_driver_block,
                                                    type_driver_block)
from dune.codegen.pdelab.driver.gridfunctionspace import (name_trial_gfs,
                                                          name_leafview,
                                                          type_leafview,
                                                          type_range,
                                                          )
from ufl import FiniteElement, MixedElement, TensorElement, VectorElement, TensorProductElement


def interpolate_dirichlet_data(name):
    element = get_trial_element()
    func = preprocess_leaf_data(element, "interpolate_expression", applyZeroDefault=False)
    if func is not None:
        bf = name_boundary_grid_function(element, func)
        gfs = name_trial_gfs()
        interpolate_vector(bf, gfs, name)


@preamble(section="vector", kernel="driver_block")
def interpolate_vector(func, gfs, name):
    return "Dune::PDELab::interpolate(*{}, *{}, *{});".format(func,
                                                              gfs,
                                                              name,
                                                              )


@preamble(section="vector", kernel="driver_block")
def define_compositegfs_parameterfunction(name, children):
    return "Dune::PDELab::CompositeGridFunction<{}> {}({});".format(', '.join('decltype({})'.format(c) for c in children),
                                                                    name,
                                                                    ', '.join(children))


@class_member(classtag="driver_block")
def typedef_boundary_grid_function(name, func):
    leafview_type = type_leafview()
    range_type = type_range()
    boundary_function_type = type_boundary_function(func)
    # palpo TODO: 1 in the format below!
    return "using {} = Dune::PDELab::LocalCallableToGridFunctionAdapter<{}, {}, {}, {}>;".format(name,
                                                                                                 leafview_type,
                                                                                                 range_type,
                                                                                                 1,
                                                                                                 boundary_function_type)


def type_boundary_grid_function(func):
    name = "BoundaryGridFunction"
    typedef_boundary_grid_function(name, func)
    return name


@class_member(classtag="driver_block")
def declare_boundary_grid_function(name, func):
    bgf_type = type_boundary_grid_function(func)
    return "std::shared_ptr<{}> {};".format(bgf_type, name)


@preamble(section="vector", kernel="driver_block")
def define_boundary_grid_function(name, func):
    declare_boundary_grid_function(name, func)
    gv = name_leafview()
    boundary_function = name_boundary_function(func)
    bgf_type = type_boundary_grid_function(func)
    include_file('dune/pdelab/function/callableadapter.hh', filetag='driver')
    if is_stationary():
        return "{} = std::make_shared<{}>({}, *{});".format(name, bgf_type, gv, boundary_function)
    else:
        # palpo TODO
        assert False
        from dune.codegen.pdelab.driver.gridoperator import name_localoperator
        lop = name_localoperator(get_form_ident())
        return "auto {} = Dune::PDELab::makeInstationaryGridFunctionFromCallable({}, {}, {});".format(name,
                                                                                                      gv,
                                                                                                      lambdaname,
                                                                                                      lop,
                                                                                                      )


@cached
def name_boundary_grid_function(element, func):
    assert isinstance(func, tuple)

    if isinstance(element, MixedElement):
        # palpo TODO
        assert False
        k = 0
        childs = []
        for subel in element.sub_elements():
            childs.append(name_boundary_grid_function(subel, func[k:k + subel.value_size()]))
            k = k + subel.value_size()
        name = "_".join(childs)
        if len(childs) == 1:
            name = "{}_dummy".format(name)
        define_compositegfs_parameterfunction(name, tuple(childs))
        return name
    else:
        assert isinstance(element, (FiniteElement, TensorProductElement))
        name = "boundary_grid_function"
        define_boundary_grid_function(name, func)
        return name


def boundary_lambda(func):
    assert isinstance(func, tuple)
    assert len(func) == 1
    func = func[0]

    if func is None:
        func = 0.0
    from ufl.classes import Expr
    if isinstance(func, (int, float)):
        return "[&](const auto& x){{ return {}; }}".format(float(func))
    elif isinstance(func, Expr):
        from dune.codegen.pdelab.driver.visitor import ufl_to_code
        return "[&](const auto& is, const auto& xl){{ {}; }}".format(ufl_to_code(func))
    else:
        raise NotImplementedError("What is this?")


@class_member(classtag="driver_block")
def typedef_boundary_function(name, func):
    assert isinstance(func, tuple)
    assert len(func) == 1
    func = func[0]

    range_type = type_range()
    leafview_type = type_leafview()
    if func is None:
        func = 0.0
    from ufl.classes import Expr
    if isinstance(func, (int, float)):
        coordinate = "typename {}::template Codim<0>::Entity::Geometry::GlobalCoordinate".format(leafview_type)
        return "using {} = std::function<{}({})>;".format(name, range_type, coordinate)
    elif isinstance(func, Expr):
        entity = "typename {}::template Codim<0>::Entity".format(leafview_type)
        coordinate = "typename {}::template Codim<0>::Entity::Geometry::LocalCoordinate".format(leafview_type)
        return "using {} = std::function<{}({}, {})>;".format(name, range_type, entity, coordinate)
    else:
        raise NotImplementedError("What is this?")


def type_boundary_function(func):
    name = "BoundaryFunction"
    typedef_boundary_function(name, func)
    return name


@class_member(classtag="driver_block")
def declare_boundary_function(name, func):
    bf_type = type_boundary_function(func)
    return "std::shared_ptr<{}> {};".format(bf_type, name)


@preamble(section="vector", kernel="driver_block")
def define_boundary_function(name, func):
    declare_boundary_function(name, func)
    bf_type = type_boundary_function(func)
    bf_lambda = boundary_lambda(func)
    return "{} = std::make_shared<{}> ({});".format(name, bf_type, bf_lambda)


@cached
def name_boundary_function(func):
    name = get_counted_variable("function")
    define_boundary_function(name, func)
    return name


@class_member(classtag="driver_block")
def driver_block_get_boundarygridfunction(element, func):
    name = name_boundary_grid_function(element, func)
    bgf_type = type_boundary_grid_function(func)
    return ["std::shared_ptr<{}> get_boundarygridfunction(){{".format(bgf_type),
            "  return {};".format(name),
            "}"]


@preamble(section="postprocessing", kernel="main")
def main_typedef_boundary_grid_function(name, func):
    driver_block_type = type_driver_block()
    bgf_type = type_boundary_grid_function(func)
    return "using {} = {}::{};".format(name, driver_block_type, bgf_type)


def main_type_boundary_grid_function(func):
    name = "BoundaryGridFunction"
    main_typedef_boundary_grid_function(name, func)
    return name


@preamble(section="postprocessing", kernel="main")
def main_define_boundar_grid_function(name, element, func):
    driver_block_name = name_driver_block()
    driver_block_get_boundarygridfunction(element, func)
    return "auto {} = {}.get_boundarygridfunction();".format(name, driver_block_name)


@cached
def main_name_boundary_grid_function(element, func):
    assert isinstance(func, tuple)

    name = "boundary_grid_function"
    main_define_boundar_grid_function(name, element, func)
    return name
